# -*- coding: utf-8 -*-

"""Module that contains some neat functions to print nice things to terminal.
"""

from typing import List, TextIO

from enum import Enum

from sys import stdout, stderr

from termcolor import colored  # pylint: disable=import-error


QUIET: bool = False
"""Set this to True to shut the mouse of prettyprint. It is not nice but you may need it sometimes.
"""


class Color(Enum):
    RED = "red"
    GREEN = "green"
    YELLOW = "yellow"
    BLUE = "blue"
    MAGENTA = "magenta"
    CYAN = "cyan"
    WHITE = "white"


def prettyprint(text: str, color: Color = Color.WHITE, out: TextIO = stdout, bold=True, dark=False,
                underline=False, blink=False, indent=0, end='\n'):
    """Awesome print.

    Args:
        text (str): your text
        color (Color, optional): color of the text. Default is white.
        out (TextIO, optional): output. Default is stdout.
        bold (bool, optional): print bold text. Default is True.
        dark (bool, optional): print dark text. Default is False.
        underline (bool, optional): underline text. Default is False.
        blink (bool, optional): the text blinks. Default is False.
        indent (int, optional): indentation level. Default is 0. Increase by 2 spaces for each level.
        end (str, optional): end - same as for standard print function. Default is '\n'.
    """

    if QUIET:
        return

    attrs: List[str] = []
    if bold:
        attrs.append("bold")
    if dark:
        attrs.append("dark")
    if underline:
        attrs.append("underline")
    if blink:
        attrs.append("blink")

    print(f"{' ' * 2 * indent}", end='')
    print(colored(text, color.value, attrs=attrs), file=out, end=end)



# -*- coding: utf-8 -*-

from typing import List, TextIO

from dataclasses import (dataclass, field)

from sys import stdout, stderr
from sys import exit

from pathlib import Path

import subprocess as sp

import yaml  # pylint: disable=import-error

import izitest.tests
from izitest.prettyprint import (prettyprint, Color)
from izitest.parser import parse_args


__all__ = [
    "discover_testsuite",
    "run_test",
    "run_testcase",
    "run_testsuite"
]


def __run_exec(path: str, args: List[str], stdin: str, timeout: int = 1) -> sp.CompletedProcess:
    """Run an executable.

    Args:
        path (str): path to the executable (can be relative to PATH env)
        args (List[str]): list of arguments (can be empty)
        stdin (str): input of the program
        timeout (int, optional): timeout (in sec). Defaults to 1

    Returns:
        sp.CompletedProcess: the completed process.

    Raises:
        sp.TimeoutExpired: if the timeout expired.
    """
    try:
        return sp.run([path] + args, input=stdin,
                      capture_output=True, text=True, timeout=timeout)
    except FileNotFoundError:
        return sp.run([Path(path).absolute()] + args, input=stdin,
                      capture_output=True, text=True, timeout=timeout)


def __fake_run_exec(expect: dict) -> sp.CompletedProcess:
    """Fake the run of an executable.

    Args:
        expect (dict): a dictionnary containing the expected outputs (stdout, stderr, retcode)

    Returns:
        sp.CompletedProcess: a completed process, as if it was run by sp.run()
    """
    fake_stdout = expect.get("stdout", '')
    fake_stderr = expect.get("stderr", '')
    fake_retcode = expect.get("retcode", 0)

    return sp.CompletedProcess('', fake_retcode, fake_stdout, fake_stderr)


def discover_testsuite(path: Path, categories: List[str] = None) -> List[Path]:
    """Returns a list of discovered tests files that match given categories/filters (if any)

    Args:
        path (Path): path of the test suite directory
        categories (List[str], optional): categories/filters - only the files that match (at least)
        one of them will be explored

    Returns:
        List[Path]: list of discovered tests files
    """

    prettyprint("Discovering test suite...", Color.GREEN, end="\n\n")

    # Must be sorted using str (string representation) to have a 'tree-like' order
    yamlfiles: List[Path] = sorted(path.rglob("*.yaml"), key=str)

    testsfiles: List[Path] = []

    if categories is None:
        testsfiles = yamlfiles
    else:
        for f in yamlfiles:
            if any(c in str(f) for c in categories):
                testsfiles.append(f)

    return testsfiles


def run_test(fn_name: str, ref: sp.CompletedProcess, test: sp.CompletedProcess) -> (bool, str):
    """Run the test named `fn_name` (located in izitest.tests).

    Args:
        fn_name (str): name of the test function
        ref (sp.CompletedProcess): reference executable
        test (sp.CompletedProcess): tested executable

    Returns:
        bool: True if the test succeed, False otherwise
        str: diff between two outputs
    """

    fn_test = getattr(izitest.tests, fn_name)
    return fn_test(ref, test)


def run_testcase(testcase: dict, exec: Path, args: List[str], ref: Path = None, refargs: List[str] = None):
    tests: List[str] = testcase.get("tests", None)
    if tests is None:
        prettyprint("No test to run...", Color.CYAN, bold=False, indent=3, end=' ')
        prettyprint("Skipped", Color.YELLOW)
        return

    add_args: str = testcase.get("args", None)
    args += add_args

    if refargs is not None:
        refargs += add_args

    stdin: str = testcase.get("stdin", '')

    expect: dict = testcase.get("expect", None)

    try:
        if expect is None:
            if ref is None:
                prettyprint("No ref executable provided...", Color.MAGENTA, bold=False, indent=3, end=' ')
                prettyprint("Skipped", Color.RED)
                return

            ref_proc = __run_exec(ref, refargs, stdin)
        else:
            ref_proc = __fake_run_exec(expect)

        test_proc: sp.CompletedProcess = __run_exec(exec, args, stdin)
    except sp.TimeoutExpired:
        prettyprint("Timeout expired!", Color.MAGENTA, bold=False, indent=2, end=' ')
        prettyprint("Skipped", Color.YELLOW)
        return

    for test in tests:
        prettyprint(test, Color.CYAN, bold=False, indent=3, end=' ')

        passed: bool
        diff: str
        passed, diff = run_test(test, ref_proc, test_proc)
        if passed:
            prettyprint("Passed", Color.GREEN)
        else:
            prettyprint("Failed", Color.RED)
            if diff != '':
                print(diff)


def run_testsuite(args):
    testfiles: List[Path] = discover_testsuite(args.testdir, args.cat)

    prettyprint("Running test suite:")
    for path in testfiles:
        testdir: str = path.parts[0]
        prettyprint(f"Running tests in {path.relative_to(testdir)}:", Color.BLUE, indent=1)
        try:
            with open(path, 'r') as f:
                tests: dict = yaml.safe_load(f.read())
        except Exception:
            prettyprint(f"Failed to load test file {path}!", Color.RED, stderr, indent=2)

        if tests is None:
            prettyprint("Empty file...", Color.CYAN, bold=False, indent=2, end=' ')
            prettyprint("Skipped", Color.YELLOW)
            continue

        for testcase in tests:
            if testcase.get("name", None) is None:
                continue

            prettyprint(f"{testcase['name']}", Color.CYAN, indent=2)

            if not testcase.get("skip", False):
                run_testcase(testcase, args.exec[0], args.exec[1:], args.ref[0], args.ref[1:])
            else:
                prettyprint("Test is disabled...", Color.CYAN, bold=False, indent=3, end=' ')
                prettyprint("Skipped", Color.YELLOW)



# -*- coding: utf-8 -*-

from typing import List, TextIO

from pathlib import Path

from argparse import (ArgumentParser, Namespace)

import izitest
from izitest.path import *

__all__ = [
    "parse_args"
]


def parse_args() -> Namespace:
    """Parse arguments.

    Also check if exec (and ref if provided) are regular executable files and testdir is a valid directory.

    Returns:
        Namespace: parsed arguments

    Raises:
        Exception: if arguments checks failed
    """

    parser = ArgumentParser(description="Easily build a test suite.", allow_abbrev=True)

    parser.add_argument("exec", type=str, help="path to the executable you want to test")

    parser.add_argument("--ref", metavar="ref", type=str, help="specify a reference executable")

    parser.add_argument("-q", "--quiet", action="store_true",
                        help="run silently, do not output ANYTHING (even errors)!")
    parser.add_argument("-d", "--testdir", metavar="dir", type=Path, default=Path("./tests"),
                        help="path to the test suite directory (default is './tests')")
    parser.add_argument("-c", "--cat", metavar="cat", nargs='+', type=str,
                        help="run only the tests of specified categories")
    parser.add_argument("-m", "--memory", action="store_true",
                        help="if set, it will check for any memory leak using valgrind (see https://valgrind.org/)")
    parser.add_argument("-r", "--report", type=Path, metavar="file", nargs='?',
                        help="generate a report (default file is './tests_report')")

    args = parser.parse_args()

    if args.quiet:
        izitest.prettyprint.QUIET = True

    args.exec = args.exec.split(' ')
    execpath: Path = Path(args.exec[0])
    check_file(execpath) and check_exec(execpath)

    if args.ref is not None:
        args.ref = args.ref.split(' ')
        refpath: Path = Path(args.ref[0])
        check_file(refpath) and check_exec(refpath)

    check_dir(args.testdir)

    return args



import setuptools

with open("README.md", "r") as f:
    readme = f.read()

setuptools.setup(
    name="izitest",
    version="1.0.0",
    author="Kenji 'Nhqml' Gaillac",
    license="GNU GPLv3",
    description="Easily run an executable against a test suite",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/nhqml/izitest",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={
        "izitest": [
            "py.typed",
        ],
    },
    platforms=[
        "Any",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'PyYAML>=5.2',
        'termcolor>=1.1.0',
    ],
)



from typing import List, TextIO

import os
from pathlib import Path

from sys import stdout, stderr

from izitest.prettyprint import (prettyprint, Color)

__all__ = [
    "check_file",
    "check_exec",
    "check_dir"
]


def check_file(path: Path) -> None:
    """Check if given path points to a regular file.

    Args:
        path (Path): path to check

    Raises:
        Exception: if not a regular file
    """
    if not path.is_file():
        prettyprint(f"{path} is not a valid file!", Color.RED, out=stderr)
        raise Exception


def check_exec(path: Path) -> None:
    """Check if given path points to a regular file.

    Args:
        path (Path): path to check

    Raises:
        Exception: if not a regular file
    """
    if not os.access(path, os.X_OK):
        prettyprint(f"{path} is not executable!", Color.RED, out=stderr)
        raise Exception


def check_dir(path: Path) -> None:
    """Check if given path points to a regular file.

    Args:
        path (Path): path to check

    Raises:
        Exception: if not a regular file
    """
    if not path.is_dir():
        prettyprint(f"{path} is not a valid directory!", Color.RED, out=stderr)
        raise Exception



# -*- coding: utf-8 -*-

"""Module that cointains all tests.

| Test functions takes two arguments:
| - **ref** the CompletedProcess of reference
| - **test** the tested process

| They return a tuple **(bool, str)**:
| - the boolean indicates if the test was successful
| - the string represents the difference of what is tested between **ref** and **test**
"""

from typing import List, TextIO

from subprocess import CompletedProcess

from difflib import unified_diff


def __diff(ref: str, test: str) -> str:
    """Computes the unified diff between two outputs.

    Args:
        ref (str): output of the reference executable
        test (str): output of the tested executable

    Returns:
        str: difference
    """
    ref_lines = ref.splitlines(keepends=True)
    test_lines = test.splitlines(keepends=True)

    diff = ''.join(unified_diff(ref_lines, test_lines, fromfile="ref", tofile="test"))

    return diff


def same_stdout(ref: CompletedProcess, test: CompletedProcess) -> (bool, str):
    """Compare stdout of the two CompletedProcess, succeed if both are identical.

    Args:
        ref (CompletedProcess): reference executable
        test (CompletedProcess): tested executable

    Returns:
        (bool, str): tuple containing return value (True for success, False otherwise)
        and the difference (empty string if they both are the same)
    """
    diff = __diff(ref.stdout, test.stdout)

    return (diff == '', diff)


def same_stderr(ref: CompletedProcess, test: CompletedProcess) -> (bool, str):
    """Compare stderr of the two CompletedProcess, succeed if both are identical.

    Args:
        ref (CompletedProcess): reference executable
        test (CompletedProcess): tested executable

    Returns:
        (bool, str): tuple containing return value (True for success, False otherwise)
        and the difference (empty string if they both are the same)
    """
    diff = __diff(ref.stderr, test.stderr)

    # Strip stderr because there is an extra '\n' at the end
    return (diff == '', diff.strip('\n'))


def same_retcode(ref: CompletedProcess, test: CompletedProcess) -> (bool, str):
    """Compare return code of the two CompletedProcess, succeed if both are identical.

    Args:
        ref (CompletedProcess): reference executable
        test (CompletedProcess): tested executable

    Returns:
        (bool, str): tuple containing return value (True for success, False otherwise)
        and the difference (empty string if they both are the same)
    """
    if ref.returncode == test.returncode:
        return (True, '')
    else:
        return (False, f"returned {test.returncode}, expected {ref.returncode}")


def same_stdout_size(ref: CompletedProcess, test: CompletedProcess) -> (bool, str):
    """Compare stdout's number of line of the two CompletedProcess, succeed if both are identical.

    Args:
        ref (CompletedProcess): reference executable
        test (CompletedProcess): tested executable

    Returns:
        (bool, str): tuple containing return value (True for success, False otherwise)
        and the difference (empty string if they both are the same)
    """
    if len(ref.stdout.splitlines()) == len(test.stdout.splitlines()):
        return (True, '')
    else:
        return (False, __diff(ref.stdout, test.stdout))


def same_stderr_size(ref: CompletedProcess, test: CompletedProcess) -> (bool, str):
    """Compare stderr's number of line of the two CompletedProcess, succeed if both are identical.

    Args:
        ref (CompletedProcess): reference executable
        test (CompletedProcess): tested executable

    Returns:
        (bool, str): tuple containing return value (True for success, False otherwise)
        and the difference (empty string if they both are the same)
    """
    if len(ref.stderr.splitlines()) == len(test.stderr.splitlines()):
        return (True, '')
    else:
        # Strip stderr because there is an extra '\n' at the end
        return (False, __diff(ref.stderr, test.stderr.strip('\n')))


def has_stdout(ref: CompletedProcess, test: CompletedProcess) -> (bool, str):
    """Test if tested executable outputed anything on stdout.

    Args:
        ref (CompletedProcess): reference executable
        test (CompletedProcess): tested executable

    Returns:
        (bool, str): tuple containing return value (True for success, False otherwise)
        and an empty string

    Note:
        An empty string is returned only to keep the same prototype for all test functions.
    """
    return (test.stdout != '', '')


def has_stderr(ref: CompletedProcess, test: CompletedProcess) -> (bool, str):
    """Test if tested executable outputed anything on stderr.

    Args:
        ref (CompletedProcess): reference executable
        test (CompletedProcess): tested executable

    Returns:
        (bool, str): tuple containing return value (True for success, False otherwise)
        and an empty string

    Note:
        An empty string is returned only to keep the same prototype for all test functions.
    """
    return (test.stderr != '', '')


def has_no_stdout(ref: CompletedProcess, test: CompletedProcess) -> (bool, str):
    """Test if tested executable outputed nothing on stdout.

    Args:
        ref (CompletedProcess): reference executable
        test (CompletedProcess): tested executable

    Returns:
        (bool, str): tuple containing return value (True for success, False otherwise)
        and an empty string

    Note:
        An empty string is returned only to keep the same prototype for all test functions.
    """
    return (test.stdout == '', '')


def has_no_stderr(ref: CompletedProcess, test: CompletedProcess) -> (bool, str):
    """Test if tested executable outputed nothing on stderr.

    Args:
        ref (CompletedProcess): reference executable
        test (CompletedProcess): tested executable

    Returns:
        (bool, str): tuple containing return value (True for success, False otherwise)
        and an empty string

    Note:
        An empty string is returned only to keep the same prototype for all test functions.
    """
    return (test.stderr == '', '')
